@@ -1,25 +1,23 @@
-from mypy.ordered_dict import OrderedDict
-import re
-import pprint
-import sys
+from __future__ import annotations
 
-from typing_extensions import Final, TYPE_CHECKING
-from typing import Dict, List, Mapping, Optional, Pattern, Set, Tuple, Callable, Any
+import pprint
+import re
+import sys
+from typing import Any, Callable, Dict, Mapping, Pattern
+from typing_extensions import Final
 
 from mypy import defaults
+from mypy.errorcodes import ErrorCode, error_codes
 from mypy.util import get_class_descriptors, replace_object_state
 
-if TYPE_CHECKING:
-    from mypy.errors import ErrorCode
-
 
 class BuildType:
-    STANDARD = 0  # type: Final
-    MODULE = 1  # type: Final
-    PROGRAM_TEXT = 2  # type: Final
+    STANDARD: Final = 0
+    MODULE: Final = 1
+    PROGRAM_TEXT: Final = 2
 
 
-PER_MODULE_OPTIONS = {
+PER_MODULE_OPTIONS: Final = {
     # Please keep this list sorted
     "allow_redefinition",
     "allow_untyped_globals",
@@ -27,6 +25,8 @@ PER_MODULE_OPTIONS = {
     "always_true",
     "check_untyped_defs",
     "debug_cache",
+    "disable_error_code",
+    "disabled_error_codes",
     "disallow_any_decorated",
     "disallow_any_explicit",
     "disallow_any_expr",
@@ -37,6 +37,8 @@ PER_MODULE_OPTIONS = {
     "disallow_untyped_calls",
     "disallow_untyped_decorators",
     "disallow_untyped_defs",
+    "enable_error_code",
+    "enabled_error_codes",
     "follow_imports",
     "follow_imports_for_stubs",
     "ignore_errors",
@@ -46,6 +48,7 @@ PER_MODULE_OPTIONS = {
     "mypyc",
     "no_implicit_optional",
     "show_none_errors",
+    "strict_concatenate",
     "strict_equality",
     "strict_optional",
     "strict_optional_whitelist",
@@ -53,11 +56,11 @@ PER_MODULE_OPTIONS = {
     "warn_return_any",
     "warn_unreachable",
     "warn_unused_ignores",
-}  # type: Final
+}
 
-OPTIONS_AFFECTING_CACHE = ((PER_MODULE_OPTIONS |
-                            {"platform", "bazel", "plugins"})
-                           - {"debug_cache"})  # type: Final
+OPTIONS_AFFECTING_CACHE: Final = (PER_MODULE_OPTIONS | {"platform", "bazel", "plugins"}) - {
+    "debug_cache"
+}
 
 
 class Options:
@@ -65,26 +68,28 @@ class Options:
 
     def __init__(self) -> None:
         # Cache for clone_for_module()
-        self._per_module_cache = None  # type: Optional[Dict[str, Options]]
+        self._per_module_cache: dict[str, Options] | None = None
 
         # -- build options --
         self.build_type = BuildType.STANDARD
-        self.python_version = sys.version_info[:2]  # type: Tuple[int, int]
+        self.python_version: tuple[int, int] = sys.version_info[:2]
         # The executable used to search for PEP 561 packages. If this is None,
         # then mypy does not search for PEP 561 packages.
-        self.python_executable = sys.executable  # type: Optional[str]
+        self.python_executable: str | None = sys.executable
         self.platform = sys.platform
-        self.custom_typing_module = None  # type: Optional[str]
-        self.custom_typeshed_dir = None  # type: Optional[str]
-        self.mypy_path = []  # type: List[str]
-        self.report_dirs = {}  # type: Dict[str, str]
+        self.custom_typing_module: str | None = None
+        self.custom_typeshed_dir: str | None = None
+        # The abspath() version of the above, we compute it once as an optimization.
+        self.abs_custom_typeshed_dir: str | None = None
+        self.mypy_path: list[str] = []
+        self.report_dirs: dict[str, str] = {}
         # Show errors in PEP 561 packages/site-packages modules
         self.no_silence_site_packages = False
         self.no_site_packages = False
         self.ignore_missing_imports = False
         # Is ignore_missing_imports set in a per-module section
         self.ignore_missing_imports_per_module = False
-        self.follow_imports = 'normal'  # normal|silent|skip|error
+        self.follow_imports = "normal"  # normal|silent|skip|error
         # Whether to respect the follow_imports setting even for stub files.
         # Intended to be used for disabling specific stubs.
         self.follow_imports_for_stubs = False
@@ -100,7 +105,7 @@ class Options:
         # top-level __init__.py to your packages.
         self.explicit_package_bases = False
         # File names, directory names or subpaths to avoid checking
-        self.exclude = ""  # type: str
+        self.exclude: list[str] = []
 
         # disallow_any options
         self.disallow_any_generics = False
@@ -161,7 +166,7 @@ class Options:
 
         # Files in which to allow strict-Optional related errors
         # TODO: Kill this in favor of show_none_errors
-        self.strict_optional_whitelist = None   # type: Optional[List[str]]
+        self.strict_optional_whitelist: list[str] | None = None
 
         # Alternate way to show/hide strict-None-checking related errors
         self.show_none_errors = True
@@ -183,41 +188,44 @@ class Options:
         # This makes 1 == '1', 1 in ['1'], and 1 is '1' errors.
         self.strict_equality = False
 
+        # Make arguments prepended via Concatenate be truly positional-only.
+        self.strict_concatenate = False
+
         # Report an error for any branches inferred to be unreachable as a result of
         # type analysis.
         self.warn_unreachable = False
 
         # Variable names considered True
-        self.always_true = []  # type: List[str]
+        self.always_true: list[str] = []
 
         # Variable names considered False
-        self.always_false = []  # type: List[str]
+        self.always_false: list[str] = []
 
         # Error codes to disable
-        self.disable_error_code = []  # type: List[str]
-        self.disabled_error_codes = set()  # type: Set[ErrorCode]
+        self.disable_error_code: list[str] = []
+        self.disabled_error_codes: set[ErrorCode] = set()
 
         # Error codes to enable
-        self.enable_error_code = []  # type: List[str]
-        self.enabled_error_codes = set()  # type: Set[ErrorCode]
+        self.enable_error_code: list[str] = []
+        self.enabled_error_codes: set[ErrorCode] = set()
 
         # Use script name instead of __main__
         self.scripts_are_modules = False
 
         # Config file name
-        self.config_file = None  # type: Optional[str]
+        self.config_file: str | None = None
 
         # A filename containing a JSON mapping from filenames to
         # mtime/size/hash arrays, used to avoid having to recalculate
         # source hashes as often.
-        self.quickstart_file = None  # type: Optional[str]
+        self.quickstart_file: str | None = None
 
         # A comma-separated list of files/directories for mypy to type check;
         # supports globbing
-        self.files = None  # type: Optional[List[str]]
+        self.files: list[str] | None = None
 
         # Write junit.xml to given file
-        self.junit_xml = None  # type: Optional[str]
+        self.junit_xml: str | None = None
 
         # Caching and incremental checking options
         self.incremental = True
@@ -236,22 +244,24 @@ class Options:
         # in modules being compiled. Not in the config file or command line.
         self.mypyc = False
 
+        # An internal flag to modify some type-checking logic while
+        # running inspections (e.g. don't expand function definitions).
+        # Not in the config file or command line.
+        self.inspections = False
+
         # Disable the memory optimization of freeing ASTs when
         # possible. This isn't exposed as a command line option
         # because it is intended for software integrating with
         # mypy. (Like mypyc.)
         self.preserve_asts = False
 
-        # PEP 612 support is a work in progress, hide it from users
-        self.wip_pep_612 = False
-
         # Paths of user plugins
-        self.plugins = []  # type: List[str]
+        self.plugins: list[str] = []
 
         # Per-module options (raw)
-        self.per_module_options = OrderedDict()  # type: OrderedDict[str, Dict[str, object]]
-        self._glob_options = []  # type: List[Tuple[str, Pattern[str]]]
-        self.unused_configs = set()  # type: Set[str]
+        self.per_module_options: dict[str, dict[str, object]] = {}
+        self._glob_options: list[tuple[str, Pattern[str]]] = []
+        self.unused_configs: set[str] = set()
 
         # -- development options --
         self.verbosity = 0  # More verbose messages (for troubleshooting)
@@ -261,6 +271,8 @@ class Options:
         self.dump_type_stats = False
         self.dump_inference_stats = False
         self.dump_build_stats = False
+        self.enable_incomplete_features = False
+        self.timing_stats: str | None = None
 
         # -- test options --
         # Stop after the semantic analysis phase
@@ -270,8 +282,9 @@ class Options:
         self.use_builtins_fixtures = False
 
         # -- experimental options --
-        self.shadow_file = None  # type: Optional[List[List[str]]]
-        self.show_column_numbers = False  # type: bool
+        self.shadow_file: list[list[str]] | None = None
+        self.show_column_numbers: bool = False
+        self.show_error_end: bool = False
         self.show_error_codes = False
         # Use soft word wrap and show trimmed source snippets with error location markers.
         self.pretty = False
@@ -286,15 +299,17 @@ class Options:
         self.export_types = False
         # List of package roots -- directories under these are packages even
         # if they don't have __init__.py.
-        self.package_root = []  # type: List[str]
-        self.cache_map = {}  # type: Dict[str, Tuple[str, str]]
+        self.package_root: list[str] = []
+        self.cache_map: dict[str, tuple[str, str]] = {}
         # Don't properly free objects on exit, just kill the current process.
-        self.fast_exit = False
+        self.fast_exit = True
+        # fast path for finding modules from source set
+        self.fast_module_lookup = False
         # Used to transform source code before parsing if not None
         # TODO: Make the type precise (AnyStr -> AnyStr)
-        self.transform_source = None  # type: Optional[Callable[[Any], Any]]
+        self.transform_source: Callable[[Any], Any] | None = None
         # Print full path to each file in the report.
-        self.show_absolute_path = False  # type: bool
+        self.show_absolute_path: bool = False
         # Install missing stub packages if True
         self.install_types = False
         # Install missing stub packages in non-interactive mode (don't prompt for
@@ -304,6 +319,8 @@ class Options:
         # skip most errors after this many messages have been reported.
         # -1 means unlimited.
         self.many_errors_threshold = defaults.MANY_ERRORS_THRESHOLD
+        # Enable recursive type aliases (currently experimental)
+        self.enable_recursive_aliases = False
 
     # To avoid breaking plugin compatibility, keep providing new_semantic_analyzer
     @property
@@ -313,18 +330,18 @@ class Options:
     def snapshot(self) -> object:
         """Produce a comparable snapshot of this Option"""
         # Under mypyc, we don't have a __dict__, so we need to do worse things.
-        d = dict(getattr(self, '__dict__', ()))
+        d = dict(getattr(self, "__dict__", ()))
         for k in get_class_descriptors(Options):
             if hasattr(self, k) and k != "new_semantic_analyzer":
                 d[k] = getattr(self, k)
         # Remove private attributes from snapshot
-        d = {k: v for k, v in d.items() if not k.startswith('_')}
+        d = {k: v for k, v in d.items() if not k.startswith("_")}
         return d
 
     def __repr__(self) -> str:
-        return 'Options({})'.format(pprint.pformat(self.snapshot()))
+        return f"Options({pprint.pformat(self.snapshot())})"
 
-    def apply_changes(self, changes: Dict[str, object]) -> 'Options':
+    def apply_changes(self, changes: dict[str, object]) -> Options:
         new_options = Options()
         # Under mypyc, we don't have a __dict__, so we need to do worse things.
         replace_object_state(new_options, self, copy_dict=True)
@@ -334,6 +351,20 @@ class Options:
             # This is the only option for which a per-module and a global
             # option sometimes beheave differently.
             new_options.ignore_missing_imports_per_module = True
+
+        # These two act as overrides, so apply them when cloning.
+        # Similar to global codes enabling overrides disabling, so we start from latter.
+        new_options.disabled_error_codes = self.disabled_error_codes.copy()
+        new_options.enabled_error_codes = self.enabled_error_codes.copy()
+        for code_str in new_options.disable_error_code:
+            code = error_codes[code_str]
+            new_options.disabled_error_codes.add(code)
+            new_options.enabled_error_codes.discard(code)
+        for code_str in new_options.enable_error_code:
+            code = error_codes[code_str]
+            new_options.enabled_error_codes.add(code)
+            new_options.disabled_error_codes.discard(code)
+
         return new_options
 
     def build_per_module_cache(self) -> None:
@@ -354,12 +385,10 @@ class Options:
         # than foo.bar.*.
         # (A section being "processed last" results in its config "winning".)
         # Unstructured glob configs are stored and are all checked for each module.
-        unstructured_glob_keys = [k for k in self.per_module_options.keys()
-                                  if '*' in k[:-1]]
-        structured_keys = [k for k in self.per_module_options.keys()
-                           if '*' not in k[:-1]]
-        wildcards = sorted(k for k in structured_keys if k.endswith('.*'))
-        concrete = [k for k in structured_keys if not k.endswith('.*')]
+        unstructured_glob_keys = [k for k in self.per_module_options.keys() if "*" in k[:-1]]
+        structured_keys = [k for k in self.per_module_options.keys() if "*" not in k[:-1]]
+        wildcards = sorted(k for k in structured_keys if k.endswith(".*"))
+        concrete = [k for k in structured_keys if not k.endswith(".*")]
 
         for glob in unstructured_glob_keys:
             self._glob_options.append((glob, self.compile_glob(glob)))
@@ -381,7 +410,7 @@ class Options:
         # they only count as used if actually used by a real module.
         self.unused_configs.update(structured_keys)
 
-    def clone_for_module(self, module: str) -> 'Options':
+    def clone_for_module(self, module: str) -> Options:
         """Create an Options object that incorporates per-module options.
 
         NOTE: Once this method is called all Options objects should be
@@ -402,9 +431,9 @@ class Options:
         # This is technically quadratic in the length of the path, but module paths
         # don't actually get all that long.
         options = self
-        path = module.split('.')
+        path = module.split(".")
         for i in range(len(path), 0, -1):
-            key = '.'.join(path[:i] + ['*'])
+            key = ".".join(path[:i] + ["*"])
             if key in self._per_module_cache:
                 self.unused_configs.discard(key)
                 options = self._per_module_cache[key]
@@ -412,7 +441,7 @@ class Options:
 
         # OK and *now* we need to look for unstructured glob matches.
         # We only do this for concrete modules, not structured wildcards.
-        if not module.endswith('.*'):
+        if not module.endswith(".*"):
             for key, pattern in self._glob_options:
                 if pattern.match(module):
                     self.unused_configs.discard(key)
@@ -428,11 +457,17 @@ class Options:
         # Compile one of the glob patterns to a regex so that '.*' can
         # match *zero or more* module sections. This means we compile
         # '.*' into '(\..*)?'.
-        parts = s.split('.')
-        expr = re.escape(parts[0]) if parts[0] != '*' else '.*'
+        parts = s.split(".")
+        expr = re.escape(parts[0]) if parts[0] != "*" else ".*"
         for part in parts[1:]:
-            expr += re.escape('.' + part) if part != '*' else r'(\..*)?'
-        return re.compile(expr + '\\Z')
+            expr += re.escape("." + part) if part != "*" else r"(\..*)?"
+        return re.compile(expr + "\\Z")
 
     def select_options_affecting_cache(self) -> Mapping[str, object]:
-        return {opt: getattr(self, opt) for opt in OPTIONS_AFFECTING_CACHE}
+        result: Dict[str, object] = {}
+        for opt in OPTIONS_AFFECTING_CACHE:
+            val = getattr(self, opt)
+            if opt in ("disabled_error_codes", "enabled_error_codes"):
+                val = sorted([code.code for code in val])
+            result[opt] = val
+        return result
