@@ -1,34 +1,47 @@
 """Utilities for processing .test files containing test case descriptions."""
 
-import os.path
+from __future__ import annotations
+
 import os
-import tempfile
+import os.path
 import posixpath
 import re
 import shutil
-from abc import abstractmethod
 import sys
+import tempfile
+from abc import abstractmethod
+from typing import Any, Iterator, NamedTuple, Pattern, Union
+from typing_extensions import Final, TypeAlias as _TypeAlias
 
 import pytest
-from typing import List, Tuple, Set, Optional, Iterator, Any, Dict, NamedTuple, Union
 
-from mypy.test.config import test_data_prefix, test_temp_dir, PREFIX
+from mypy.test.config import PREFIX, test_data_prefix, test_temp_dir
 
 root_dir = os.path.normpath(PREFIX)
 
+# Debuggers that we support for debugging mypyc run tests
+# implementation of using each of these debuggers is in test_run.py
+# TODO: support more debuggers
+SUPPORTED_DEBUGGERS: Final = ["gdb", "lldb"]
+
+
 # File modify/create operation: copy module contents from source_path.
-UpdateFile = NamedTuple('UpdateFile', [('module', str),
-                                       ('source_path', str),
-                                       ('target_path', str)])
+class UpdateFile(NamedTuple):
+    module: str
+    content: str
+    target_path: str
+
 
 # File delete operation: delete module file.
-DeleteFile = NamedTuple('DeleteFile', [('module', str),
-                                       ('path', str)])
-
-FileOperation = Union[UpdateFile, DeleteFile]
+class DeleteFile(NamedTuple):
+    module: str
+    path: str
 
 
-def parse_test_case(case: 'DataDrivenTestCase') -> None:
+FileOperation: _TypeAlias = Union[UpdateFile, DeleteFile]
+
+
+def parse_test_case(case: DataDrivenTestCase) -> None:
     """Parse and prepare a single case from suite with test case descriptions.
 
     This method is part of the setup phase, just before the test case is run.
@@ -43,67 +56,69 @@ def parse_test_case(case: 'DataDrivenTestCase') -> None:
     out_section_missing = case.suite.required_out_section
     normalize_output = True
 
-    files = []  # type: List[Tuple[str, str]] # path and contents
-    output_files = []  # type: List[Tuple[str, str]] # path and contents for output files
-    output = []  # type: List[str]  # Regular output errors
-    output2 = {}  # type: Dict[int, List[str]]  # Output errors for incremental, runs 2+
-    deleted_paths = {}  # type: Dict[int, Set[str]]  # from run number of paths
-    stale_modules = {}  # type: Dict[int, Set[str]]  # from run number to module names
-    rechecked_modules = {}  # type: Dict[ int, Set[str]]  # from run number module names
-    triggered = []  # type: List[str]  # Active triggers (one line per incremental step)
-    targets = {}  # type: Dict[int, List[str]]  # Fine-grained targets (per fine-grained update)
+    files: list[tuple[str, str]] = []  # path and contents
+    output_files: list[tuple[str, str | Pattern[str]]] = []  # output path and contents
+    output: list[str] = []  # Regular output errors
+    output2: dict[int, list[str]] = {}  # Output errors for incremental, runs 2+
+    deleted_paths: dict[int, set[str]] = {}  # from run number of paths
+    stale_modules: dict[int, set[str]] = {}  # from run number to module names
+    rechecked_modules: dict[int, set[str]] = {}  # from run number module names
+    triggered: list[str] = []  # Active triggers (one line per incremental step)
+    targets: dict[int, list[str]] = {}  # Fine-grained targets (per fine-grained update)
 
     # Process the parsed items. Each item has a header of form [id args],
     # optionally followed by lines of text.
     item = first_item = test_items[0]
     for item in test_items[1:]:
-        if item.id == 'file' or item.id == 'outfile':
+        if item.id in {"file", "outfile", "outfile-re"}:
             # Record an extra file needed for the test case.
             assert item.arg is not None
-            contents = expand_variables('\n'.join(item.data))
+            contents = expand_variables("\n".join(item.data))
             file_entry = (join(base_path, item.arg), contents)
-            if item.id == 'file':
+            if item.id == "file":
                 files.append(file_entry)
+            elif item.id == "outfile-re":
+                output_files.append((file_entry[0], re.compile(file_entry[1].rstrip(), re.S)))
             else:
                 output_files.append(file_entry)
-        elif item.id in ('builtins', 'builtins_py2'):
+        elif item.id in ("builtins", "builtins_py2"):
             # Use an alternative stub file for the builtins module.
             assert item.arg is not None
             mpath = join(os.path.dirname(case.file), item.arg)
-            fnam = 'builtins.pyi' if item.id == 'builtins' else '__builtin__.pyi'
-            with open(mpath, encoding='utf8') as f:
+            fnam = "builtins.pyi" if item.id == "builtins" else "__builtin__.pyi"
+            with open(mpath, encoding="utf8") as f:
                 files.append((join(base_path, fnam), f.read()))
-        elif item.id == 'typing':
+        elif item.id == "typing":
             # Use an alternative stub file for the typing module.
             assert item.arg is not None
             src_path = join(os.path.dirname(case.file), item.arg)
-            with open(src_path, encoding='utf8') as f:
-                files.append((join(base_path, 'typing.pyi'), f.read()))
-        elif re.match(r'stale[0-9]*$', item.id):
-            passnum = 1 if item.id == 'stale' else int(item.id[len('stale'):])
+            with open(src_path, encoding="utf8") as f:
+                files.append((join(base_path, "typing.pyi"), f.read()))
+        elif re.match(r"stale[0-9]*$", item.id):
+            passnum = 1 if item.id == "stale" else int(item.id[len("stale") :])
             assert passnum > 0
-            modules = (set() if item.arg is None else {t.strip() for t in item.arg.split(',')})
+            modules = set() if item.arg is None else {t.strip() for t in item.arg.split(",")}
             stale_modules[passnum] = modules
-        elif re.match(r'rechecked[0-9]*$', item.id):
-            passnum = 1 if item.id == 'rechecked' else int(item.id[len('rechecked'):])
+        elif re.match(r"rechecked[0-9]*$", item.id):
+            passnum = 1 if item.id == "rechecked" else int(item.id[len("rechecked") :])
             assert passnum > 0
-            modules = (set() if item.arg is None else {t.strip() for t in item.arg.split(',')})
+            modules = set() if item.arg is None else {t.strip() for t in item.arg.split(",")}
             rechecked_modules[passnum] = modules
-        elif re.match(r'targets[0-9]*$', item.id):
-            passnum = 1 if item.id == 'targets' else int(item.id[len('targets'):])
+        elif re.match(r"targets[0-9]*$", item.id):
+            passnum = 1 if item.id == "targets" else int(item.id[len("targets") :])
             assert passnum > 0
-            reprocessed = [] if item.arg is None else [t.strip() for t in item.arg.split(',')]
+            reprocessed = [] if item.arg is None else [t.strip() for t in item.arg.split(",")]
             targets[passnum] = reprocessed
-        elif item.id == 'delete':
+        elif item.id == "delete":
             # File/directory to delete during a multi-step test case
             assert item.arg is not None
-            m = re.match(r'(.*)\.([0-9]+)$', item.arg)
-            assert m, 'Invalid delete section: {}'.format(item.arg)
+            m = re.match(r"(.*)\.([0-9]+)$", item.arg)
+            assert m, f"Invalid delete section: {item.arg}"
             num = int(m.group(2))
-            assert num >= 2, "Can't delete during step {}".format(num)
+            assert num >= 2, f"Can't delete during step {num}"
             full = join(base_path, m.group(1))
             deleted_paths.setdefault(num, set()).add(full)
-        elif re.match(r'out[0-9]*$', item.id):
+        elif re.match(r"out[0-9]*$", item.id):
             if item.arg is None:
                 args = []
             else:
@@ -111,12 +126,13 @@ def parse_test_case(case: 'DataDrivenTestCase') -> None:
 
             version_check = True
             for arg in args:
-                if arg == 'skip-path-normalization':
+                if arg == "skip-path-normalization":
                     normalize_output = False
                 if arg.startswith("version"):
-                    if arg[7:9] != ">=":
+                    compare_op = arg[7:9]
+                    if compare_op not in {">=", "=="}:
                         raise ValueError(
-                            "{}, line {}: Only >= version checks are currently supported".format(
+                            "{}, line {}: Only >= and == version checks are currently supported".format(
                                 case.file, item.line
                             )
                         )
@@ -126,49 +142,61 @@ def parse_test_case(case: 'DataDrivenTestCase') -> None:
                     except ValueError:
                         raise ValueError(
                             '{}, line {}: "{}" is not a valid python version'.format(
-                                case.file, item.line, version_str))
-                    if not sys.version_info >= version:
-                        version_check = False
-
+                                case.file, item.line, version_str
+                            )
+                        )
+                    if compare_op == ">=":
+                        version_check = sys.version_info >= version
+                    elif compare_op == "==":
+                        if not 1 < len(version) < 4:
+                            raise ValueError(
+                                "{}, line {}: Only minor or patch version checks "
+                                'are currently supported with "==": "{}"'.format(
+                                    case.file, item.line, version_str
+                                )
+                            )
+                        version_check = sys.version_info[: len(version)] == version
             if version_check:
                 tmp_output = [expand_variables(line) for line in item.data]
-                if os.path.sep == '\\' and normalize_output:
+                if os.path.sep == "\\" and normalize_output:
                     tmp_output = [fix_win_path(line) for line in tmp_output]
-                if item.id == 'out' or item.id == 'out1':
+                if item.id == "out" or item.id == "out1":
                     output = tmp_output
                 else:
-                    passnum = int(item.id[len('out'):])
+                    passnum = int(item.id[len("out") :])
                     assert passnum > 1
                     output2[passnum] = tmp_output
                 out_section_missing = False
-        elif item.id == 'triggered' and item.arg is None:
+        elif item.id == "triggered" and item.arg is None:
             triggered = item.data
         else:
             raise ValueError(
-                'Invalid section header {} in {} at line {}'.format(
-                    item.id, case.file, item.line))
+                f"Invalid section header {item.id} in {case.file} at line {item.line}"
+            )
 
     if out_section_missing:
-        raise ValueError(
-            '{}, line {}: Required output section not found'.format(
-                case.file, first_item.line))
+        raise ValueError(f"{case.file}, line {first_item.line}: Required output section not found")
 
     for passnum in stale_modules.keys():
         if passnum not in rechecked_modules:
             # If the set of rechecked modules isn't specified, make it the same as the set
             # of modules with a stale public interface.
             rechecked_modules[passnum] = stale_modules[passnum]
-        if (passnum in stale_modules
-                and passnum in rechecked_modules
-                and not stale_modules[passnum].issubset(rechecked_modules[passnum])):
+        if (
+            passnum in stale_modules
+            and passnum in rechecked_modules
+            and not stale_modules[passnum].issubset(rechecked_modules[passnum])
+        ):
             raise ValueError(
-                ('Stale modules after pass {} must be a subset of rechecked '
-                 'modules ({}:{})').format(passnum, case.file, first_item.line))
+                (
+                    "Stale modules after pass {} must be a subset of rechecked " "modules ({}:{})"
+                ).format(passnum, case.file, first_item.line)
+            )
 
     input = first_item.data
-    expand_errors(input, output, 'main')
+    expand_errors(input, output, "main")
     for file_path, contents in files:
-        expand_errors(contents.split('\n'), output, file_path)
+        expand_errors(contents.split("\n"), output, file_path)
 
     case.input = input
     case.output = output
@@ -188,71 +216,81 @@ class DataDrivenTestCase(pytest.Item):
     """Holds parsed data-driven test cases, and handles directory setup and teardown."""
 
     # Override parent member type
-    parent = None  # type: DataSuiteCollector
+    parent: DataSuiteCollector
 
-    input = None  # type: List[str]
-    output = None  # type: List[str]  # Output for the first pass
-    output2 = None  # type: Dict[int, List[str]]  # Output for runs 2+, indexed by run number
+    input: list[str]
+    output: list[str]  # Output for the first pass
+    output2: dict[int, list[str]]  # Output for runs 2+, indexed by run number
 
     # full path of test suite
-    file = ''
+    file = ""
     line = 0
 
     # (file path, file content) tuples
-    files = None  # type: List[Tuple[str, str]]
-    expected_stale_modules = None  # type: Dict[int, Set[str]]
-    expected_rechecked_modules = None  # type: Dict[int, Set[str]]
-    expected_fine_grained_targets = None  # type: Dict[int, List[str]]
+    files: list[tuple[str, str]]
+    expected_stale_modules: dict[int, set[str]]
+    expected_rechecked_modules: dict[int, set[str]]
+    expected_fine_grained_targets: dict[int, list[str]]
 
     # Whether or not we should normalize the output to standardize things like
     # forward vs backward slashes in file paths for Windows vs Linux.
     normalize_output = True
 
     # Extra attributes used by some tests.
-    last_line = None  # type: int
-    output_files = None  # type: List[Tuple[str, str]] # Path and contents for output files
-    deleted_paths = None  # type: Dict[int, Set[str]]  # Mapping run number -> paths
-    triggered = None  # type: List[str]  # Active triggers (one line per incremental step)
+    last_line: int
+    output_files: list[tuple[str, str | Pattern[str]]]  # Path and contents for output files
+    deleted_paths: dict[int, set[str]]  # Mapping run number -> paths
+    triggered: list[str]  # Active triggers (one line per incremental step)
 
-    def __init__(self,
-                 parent: 'DataSuiteCollector',
-                 suite: 'DataSuite',
-                 file: str,
-                 name: str,
-                 writescache: bool,
-                 only_when: str,
-                 platform: Optional[str],
-                 skip: bool,
-                 data: str,
-                 line: int) -> None:
+    def __init__(
+        self,
+        parent: DataSuiteCollector,
+        suite: DataSuite,
+        file: str,
+        name: str,
+        writescache: bool,
+        only_when: str,
+        platform: str | None,
+        skip: bool,
+        xfail: bool,
+        data: str,
+        line: int,
+    ) -> None:
         super().__init__(name, parent)
         self.suite = suite
         self.file = file
         self.writescache = writescache
         self.only_when = only_when
-        if ((platform == 'windows' and sys.platform != 'win32')
-                or (platform == 'posix' and sys.platform == 'win32')):
+        if (platform == "windows" and sys.platform != "win32") or (
+            platform == "posix" and sys.platform == "win32"
+        ):
             skip = True
         self.skip = skip
+        self.xfail = xfail
         self.data = data
         self.line = line
-        self.old_cwd = None  # type: Optional[str]
-        self.tmpdir = None  # type: Optional[tempfile.TemporaryDirectory[str]]
+        self.old_cwd: str | None = None
+        self.tmpdir: tempfile.TemporaryDirectory[str] | None = None
 
     def runtest(self) -> None:
         if self.skip:
             pytest.skip()
-        suite = self.parent.obj()
+        # TODO: add a better error message for when someone uses skip and xfail at the same time
+        elif self.xfail:
+            self.add_marker(pytest.mark.xfail)
+        parent = self.getparent(DataSuiteCollector)
+        assert parent is not None, "Should not happen"
+        suite = parent.obj()
         suite.setup()
         try:
             suite.run_case(self)
         except Exception:
             # As a debugging aid, support copying the contents of the tmp directory somewhere
-            save_dir = self.config.getoption('--save-failures-to', None)  # type: Optional[str]
+            save_dir: str | None = self.config.getoption("--save-failures-to", None)
             if save_dir:
                 assert self.tmpdir is not None
                 target_dir = os.path.join(save_dir, os.path.basename(self.tmpdir.name))
-                print("Copying data from test {} to {}".format(self.name, target_dir))
+                print(f"Copying data from test {self.name} to {target_dir}")
                 if not os.path.isabs(target_dir):
                     assert self.old_cwd
                     target_dir = os.path.join(self.old_cwd, target_dir)
@@ -262,18 +300,41 @@ class DataDrivenTestCase(pytest.Item):
     def setup(self) -> None:
         parse_test_case(case=self)
         self.old_cwd = os.getcwd()
-        self.tmpdir = tempfile.TemporaryDirectory(prefix='mypy-test-')
+        self.tmpdir = tempfile.TemporaryDirectory(prefix="mypy-test-")
         os.chdir(self.tmpdir.name)
         os.mkdir(test_temp_dir)
+
+        # Precalculate steps for find_steps()
+        steps: dict[int, list[FileOperation]] = {}
+
         for path, content in self.files:
-            dir = os.path.dirname(path)
-            os.makedirs(dir, exist_ok=True)
-            with open(path, 'w', encoding='utf8') as f:
-                f.write(content)
+            m = re.match(r".*\.([0-9]+)$", path)
+            if m:
+                # Skip writing subsequent incremental steps - rather
+                # store them as operations.
+                num = int(m.group(1))
+                assert num >= 2
+                target_path = re.sub(r"\.[0-9]+$", "", path)
+                module = module_from_path(target_path)
+                operation = UpdateFile(module, content, target_path)
+                steps.setdefault(num, []).append(operation)
+            else:
+                # Write the first incremental steps
+                dir = os.path.dirname(path)
+                os.makedirs(dir, exist_ok=True)
+                with open(path, "w", encoding="utf8") as f:
+                    f.write(content)
+
+        for num, paths in self.deleted_paths.items():
+            assert num >= 2
+            for path in paths:
+                module = module_from_path(path)
+                steps.setdefault(num, []).append(DeleteFile(module, path))
+        max_step = max(steps) if steps else 2
+        self.steps = [steps.get(num, []) for num in range(2, max_step + 1)]
 
     def teardown(self) -> None:
-        assert self.old_cwd is not None and self.tmpdir is not None, \
-            "test was not properly set up"
+        assert self.old_cwd is not None and self.tmpdir is not None, "test was not properly set up"
         os.chdir(self.old_cwd)
         try:
             self.tmpdir.cleanup()
@@ -282,10 +343,10 @@ class DataDrivenTestCase(pytest.Item):
         self.old_cwd = None
         self.tmpdir = None
 
-    def reportinfo(self) -> Tuple[str, int, str]:
+    def reportinfo(self) -> tuple[str, int, str]:
         return self.file, self.line, self.name
 
-    def repr_failure(self, excinfo: Any, style: Optional[Any] = None) -> str:
+    def repr_failure(self, excinfo: Any, style: Any | None = None) -> str:
         if excinfo.errisinstance(SystemExit):
             # We assume that before doing exit() (which raises SystemExit) we've printed
             # enough context about what happened so that a stack trace is not useful.
@@ -294,11 +355,11 @@ class DataDrivenTestCase(pytest.Item):
             excrepr = excinfo.exconly()
         else:
             self.parent._prunetraceback(excinfo)
-            excrepr = excinfo.getrepr(style='short')
+            excrepr = excinfo.getrepr(style="short")
 
-        return "data: {}:{}:\n{}".format(self.file, self.line, excrepr)
+        return f"data: {self.file}:{self.line}:\n{excrepr}"
 
-    def find_steps(self) -> List[List[FileOperation]]:
+    def find_steps(self) -> list[list[FileOperation]]:
         """Return a list of descriptions of file operations for each incremental step.
 
         The first list item corresponds to the first incremental step, the second for the
@@ -307,33 +368,16 @@ class DataDrivenTestCase(pytest.Item):
 
         Defaults to having two steps if there aern't any operations.
         """
-        steps = {}  # type: Dict[int, List[FileOperation]]
-        for path, _ in self.files:
-            m = re.match(r'.*\.([0-9]+)$', path)
-            if m:
-                num = int(m.group(1))
-                assert num >= 2
-                target_path = re.sub(r'\.[0-9]+$', '', path)
-                module = module_from_path(target_path)
-                operation = UpdateFile(module, path, target_path)
-                steps.setdefault(num, []).append(operation)
-        for num, paths in self.deleted_paths.items():
-            assert num >= 2
-            for path in paths:
-                module = module_from_path(path)
-                steps.setdefault(num, []).append(DeleteFile(module, path))
-        max_step = max(steps) if steps else 2
-        return [steps.get(num, []) for num in range(2, max_step + 1)]
+        return self.steps
 
 
 def module_from_path(path: str) -> str:
-    path = re.sub(r'\.pyi?$', '', path)
+    path = re.sub(r"\.pyi?$", "", path)
     # We can have a mix of Unix-style and Windows-style separators.
-    parts = re.split(r'[/\\]', path)
-    assert parts[0] == test_temp_dir
+    parts = re.split(r"[/\\]", path)
     del parts[0]
-    module = '.'.join(parts)
-    module = re.sub(r'\.__init__$', '', module)
+    module = ".".join(parts)
+    module = re.sub(r"\.__init__$", "", module)
     return module
 
 
@@ -345,39 +389,38 @@ class TestItem:
       .. data ..
     """
 
-    id = ''
-    arg = ''  # type: Optional[str]
+    id = ""
+    arg: str | None = ""
 
     # Text data, array of 8-bit strings
-    data = None  # type: List[str]
+    data: list[str]
 
-    file = ''
+    file = ""
     line = 0  # Line number in file
 
-    def __init__(self, id: str, arg: Optional[str], data: List[str],
-                 line: int) -> None:
+    def __init__(self, id: str, arg: str | None, data: list[str], line: int) -> None:
         self.id = id
         self.arg = arg
         self.data = data
         self.line = line
 
 
-def parse_test_data(raw_data: str, name: str) -> List[TestItem]:
+def parse_test_data(raw_data: str, name: str) -> list[TestItem]:
     """Parse a list of lines that represent a sequence of test items."""
 
-    lines = ['', '[case ' + name + ']'] + raw_data.split('\n')
-    ret = []  # type: List[TestItem]
-    data = []  # type: List[str]
+    lines = ["", "[case " + name + "]"] + raw_data.split("\n")
+    ret: list[TestItem] = []
+    data: list[str] = []
 
-    id = None  # type: Optional[str]
-    arg = None  # type: Optional[str]
+    id: str | None = None
+    arg: str | None = None
 
     i = 0
     i0 = 0
     while i < len(lines):
         s = lines[i].strip()
 
-        if lines[i].startswith('[') and s.endswith(']'):
+        if lines[i].startswith("[") and s.endswith("]"):
             if id:
                 data = collapse_line_continuation(data)
                 data = strip_list(data)
@@ -386,15 +429,15 @@ def parse_test_data(raw_data: str, name: str) -> List[TestItem]:
             i0 = i
             id = s[1:-1]
             arg = None
-            if ' ' in id:
-                arg = id[id.index(' ') + 1:]
-                id = id[:id.index(' ')]
+            if " " in id:
+                arg = id[id.index(" ") + 1 :]
+                id = id[: id.index(" ")]
             data = []
-        elif lines[i].startswith('\\['):
+        elif lines[i].startswith("\\["):
             data.append(lines[i][1:])
-        elif not lines[i].startswith('--'):
+        elif not lines[i].startswith("--"):
             data.append(lines[i])
-        elif lines[i].startswith('----'):
+        elif lines[i].startswith("----"):
             data.append(lines[i][2:])
         i += 1
 
@@ -407,42 +450,42 @@ def parse_test_data(raw_data: str, name: str) -> List[TestItem]:
     return ret
 
 
-def strip_list(l: List[str]) -> List[str]:
+def strip_list(l: list[str]) -> list[str]:
     """Return a stripped copy of l.
 
     Strip whitespace at the end of all lines, and strip all empty
     lines from the end of the array.
     """
 
-    r = []  # type: List[str]
+    r: list[str] = []
     for s in l:
         # Strip spaces at end of line
-        r.append(re.sub(r'\s+$', '', s))
+        r.append(re.sub(r"\s+$", "", s))
 
-    while len(r) > 0 and r[-1] == '':
+    while len(r) > 0 and r[-1] == "":
         r.pop()
 
     return r
 
 
-def collapse_line_continuation(l: List[str]) -> List[str]:
-    r = []  # type: List[str]
+def collapse_line_continuation(l: list[str]) -> list[str]:
+    r: list[str] = []
     cont = False
     for s in l:
-        ss = re.sub(r'\\$', '', s)
+        ss = re.sub(r"\\$", "", s)
         if cont:
-            r[-1] += re.sub('^ +', '', ss)
+            r[-1] += re.sub("^ +", "", ss)
         else:
             r.append(ss)
-        cont = s.endswith('\\')
+        cont = s.endswith("\\")
     return r
 
 
 def expand_variables(s: str) -> str:
-    return s.replace('<ROOT>', root_dir)
+    return s.replace("<ROOT>", root_dir)
 
 
-def expand_errors(input: List[str], output: List[str], fnam: str) -> None:
+def expand_errors(input: list[str], output: list[str], fnam: str) -> None:
     """Transform comments such as '# E: message' or
     '# E:3: message' in input.
 
@@ -451,26 +494,24 @@ def expand_errors(input: List[str], output: List[str], fnam: str) -> None:
 
     for i in range(len(input)):
         # The first in the split things isn't a comment
-        for possible_err_comment in input[i].split(' # ')[1:]:
+        for possible_err_comment in input[i].split(" # ")[1:]:
             m = re.search(
-                r'^([ENW]):((?P<col>\d+):)? (?P<message>.*)$',
-                possible_err_comment.strip())
+                r"^([ENW]):((?P<col>\d+):)? (?P<message>.*)$", possible_err_comment.strip()
+            )
             if m:
-                if m.group(1) == 'E':
-                    severity = 'error'
-                elif m.group(1) == 'N':
-                    severity = 'note'
-                elif m.group(1) == 'W':
-                    severity = 'warning'
-                col = m.group('col')
-                message = m.group('message')
-                message = message.replace('\\#', '#')  # adds back escaped # character
+                if m.group(1) == "E":
+                    severity = "error"
+                elif m.group(1) == "N":
+                    severity = "note"
+                elif m.group(1) == "W":
+                    severity = "warning"
+                col = m.group("col")
+                message = m.group("message")
+                message = message.replace("\\#", "#")  # adds back escaped # character
                 if col is None:
-                    output.append(
-                        '{}:{}: {}: {}'.format(fnam, i + 1, severity, message))
+                    output.append(f"{fnam}:{i + 1}: {severity}: {message}")
                 else:
-                    output.append('{}:{}:{}: {}: {}'.format(
-                        fnam, i + 1, col, severity, message))
+                    output.append(f"{fnam}:{i + 1}:{col}: {severity}: {message}")
 
 
 def fix_win_path(line: str) -> str:
@@ -478,14 +519,13 @@ def fix_win_path(line: str) -> str:
 
     E.g. foo\bar.py -> foo/bar.py.
     """
-    line = line.replace(root_dir, root_dir.replace('\\', '/'))
-    m = re.match(r'^([\S/]+):(\d+:)?(\s+.*)', line)
+    line = line.replace(root_dir, root_dir.replace("\\", "/"))
+    m = re.match(r"^([\S/]+):(\d+:)?(\s+.*)", line)
     if not m:
         return line
     else:
         filename, lineno, message = m.groups()
-        return '{}:{}{}'.format(filename.replace('\\', '/'),
-                                lineno or '', message)
+        return "{}:{}{}".format(filename.replace("\\", "/"), lineno or "", message)
 
 
 def fix_cobertura_filename(line: str) -> str:
@@ -496,9 +536,9 @@ def fix_cobertura_filename(line: str) -> str:
     m = re.search(r'<class .* filename="(?P<filename>.*?)"', line)
     if not m:
         return line
-    return '{}{}{}'.format(line[:m.start(1)],
-                           m.group('filename').replace('\\', '/'),
-                           line[m.end(1):])
+    return "{}{}{}".format(
+        line[: m.start(1)], m.group("filename").replace("\\", "/"), line[m.end(1) :]
+    )
 
 
 ##
@@ -511,22 +551,39 @@ def fix_cobertura_filename(line: str) -> str:
 # This function name is special to pytest.  See
 # https://docs.pytest.org/en/latest/reference.html#initialization-hooks
 def pytest_addoption(parser: Any) -> None:
-    group = parser.getgroup('mypy')
-    group.addoption('--update-data', action='store_true', default=False,
-                    help='Update test data to reflect actual output'
-                         ' (supported only for certain tests)')
-    group.addoption('--save-failures-to', default=None,
-                    help='Copy the temp directories from failing tests to a target directory')
-    group.addoption('--mypy-verbose', action='count',
-                    help='Set the verbose flag when creating mypy Options')
-    group.addoption('--mypyc-showc', action='store_true', default=False,
-                    help='Display C code on mypyc test failures')
+    group = parser.getgroup("mypy")
+    group.addoption(
+        "--update-data",
+        action="store_true",
+        default=False,
+        help="Update test data to reflect actual output (supported only for certain tests)",
+    )
+    group.addoption(
+        "--save-failures-to",
+        default=None,
+        help="Copy the temp directories from failing tests to a target directory",
+    )
+    group.addoption(
+        "--mypy-verbose", action="count", help="Set the verbose flag when creating mypy Options"
+    )
+    group.addoption(
+        "--mypyc-showc",
+        action="store_true",
+        default=False,
+        help="Display C code on mypyc test failures",
+    )
+    group.addoption(
+        "--mypyc-debug",
+        default=None,
+        dest="debugger",
+        choices=SUPPORTED_DEBUGGERS,
+        help="Run the first mypyc run test with the specified debugger",
+    )
 
 
 # This function name is special to pytest.  See
 # http://doc.pytest.org/en/latest/writing_plugins.html#collection-hooks
-def pytest_pycollect_makeitem(collector: Any, name: str,
-                              obj: object) -> 'Optional[Any]':
+def pytest_pycollect_makeitem(collector: Any, name: str, obj: object) -> Any | None:
     """Called by pytest on each object in modules configured in conftest.py files.
 
     collector is pytest.Collector, returns Optional[pytest.Class]
@@ -543,26 +600,39 @@ def pytest_pycollect_makeitem(collector: Any, name: str,
     return None
 
 
-def split_test_cases(parent: 'DataSuiteCollector', suite: 'DataSuite',
-                     file: str) -> Iterator['DataDrivenTestCase']:
+def split_test_cases(
+    parent: DataFileCollector, suite: DataSuite, file: str
+) -> Iterator[DataDrivenTestCase]:
     """Iterate over raw test cases in file, at collection time, ignoring sub items.
 
     The collection phase is slow, so any heavy processing should be deferred to after
     uninteresting tests are filtered (when using -k PATTERN switch).
     """
-    with open(file, encoding='utf-8') as f:
+    with open(file, encoding="utf-8") as f:
         data = f.read()
-    cases = re.split(r'^\[case ([a-zA-Z_0-9]+)'
-                     r'(-writescache)?'
-                     r'(-only_when_cache|-only_when_nocache)?'
-                     r'(-posix|-windows)?'
-                     r'(-skip)?'
-                     r'\][ \t]*$\n',
-                     data,
-                     flags=re.DOTALL | re.MULTILINE)
-    line_no = cases[0].count('\n') + 1
-    for i in range(1, len(cases), 6):
-        name, writescache, only_when, platform_flag, skip, data = cases[i:i + 6]
+    # number of groups in the below regex
+    NUM_GROUPS = 7
+    cases = re.split(
+        r"^\[case ([a-zA-Z_0-9]+)"
+        r"(-writescache)?"
+        r"(-only_when_cache|-only_when_nocache)?"
+        r"(-posix|-windows)?"
+        r"(-skip)?"
+        r"(-xfail)?"
+        r"\][ \t]*$\n",
+        data,
+        flags=re.DOTALL | re.MULTILINE,
+    )
+    line_no = cases[0].count("\n") + 1
+    test_names = set()
+    for i in range(1, len(cases), NUM_GROUPS):
+        name, writescache, only_when, platform_flag, skip, xfail, data = cases[i : i + NUM_GROUPS]
+        if name in test_names:
+            raise RuntimeError(
+                'Found a duplicate test name "{}" in {} on line {}'.format(
+                    name, parent.name, line_no
+                )
+            )
         platform = platform_flag[1:] if platform_flag else None
         yield DataDrivenTestCase.from_parent(
             parent=parent,
@@ -573,51 +643,82 @@ def split_test_cases(parent: 'DataSuiteCollector', suite: 'DataSuite',
             only_when=only_when,
             platform=platform,
             skip=bool(skip),
+            xfail=bool(xfail),
             data=data,
             line=line_no,
         )
-        line_no += data.count('\n') + 1
+        line_no += data.count("\n") + 1
+
+        # Record existing tests to prevent duplicates:
+        test_names.update({name})
 
 
 class DataSuiteCollector(pytest.Class):
-    def collect(self) -> Iterator[pytest.Item]:
+    def collect(self) -> Iterator[DataFileCollector]:
         """Called by pytest on each of the object returned from pytest_pycollect_makeitem"""
 
         # obj is the object for which pytest_pycollect_makeitem returned self.
-        suite = self.obj  # type: DataSuite
-        for f in suite.files:
-            yield from split_test_cases(self, suite, os.path.join(suite.data_prefix, f))
+        suite: DataSuite = self.obj
+
+        assert os.path.isdir(
+            suite.data_prefix
+        ), f"Test data prefix ({suite.data_prefix}) not set correctly"
+
+        for data_file in suite.files:
+            yield DataFileCollector.from_parent(parent=self, name=data_file)
+
+
+class DataFileCollector(pytest.Collector):
+    """Represents a single `.test` data driven test file.
+
+    More context: https://github.com/python/mypy/issues/11662
+    """
+
+    parent: DataSuiteCollector
+
+    @classmethod  # We have to fight with pytest here:
+    def from_parent(  # type: ignore[override]
+        cls, parent: DataSuiteCollector, *, name: str
+    ) -> DataFileCollector:
+        return super().from_parent(parent, name=name)
+
+    def collect(self) -> Iterator[DataDrivenTestCase]:
+        yield from split_test_cases(
+            parent=self,
+            suite=self.parent.obj,
+            file=os.path.join(self.parent.obj.data_prefix, self.name),
+        )
 
 
 def add_test_name_suffix(name: str, suffix: str) -> str:
     # Find magic suffix of form "-foobar" (used for things like "-skip").
-    m = re.search(r'-[-A-Za-z0-9]+$', name)
+    m = re.search(r"-[-A-Za-z0-9]+$", name)
     if m:
         # Insert suite-specific test name suffix before the magic suffix
         # which must be the last thing in the test case name since we
         # are using endswith() checks.
         magic_suffix = m.group(0)
-        return name[:-len(magic_suffix)] + suffix + magic_suffix
+        return name[: -len(magic_suffix)] + suffix + magic_suffix
     else:
         return name + suffix
 
 
 def is_incremental(testcase: DataDrivenTestCase) -> bool:
-    return 'incremental' in testcase.name.lower() or 'incremental' in testcase.file
+    return "incremental" in testcase.name.lower() or "incremental" in testcase.file
 
 
 def has_stable_flags(testcase: DataDrivenTestCase) -> bool:
-    if any(re.match(r'# flags[2-9]:', line) for line in testcase.input):
+    if any(re.match(r"# flags[2-9]:", line) for line in testcase.input):
         return False
     for filename, contents in testcase.files:
-        if os.path.basename(filename).startswith('mypy.ini.'):
+        if os.path.basename(filename).startswith("mypy.ini."):
             return False
     return True
 
 
 class DataSuite:
     # option fields - class variables
-    files = None  # type: List[str]
+    files: list[str]
 
     base_path = test_temp_dir
 
@@ -630,11 +731,10 @@ class DataSuite:
 
     # Name suffix automatically added to each test case in the suite (can be
     # used to distinguish test cases in suites that share data files)
-    test_name_suffix = ''
+    test_name_suffix = ""
 
     def setup(self) -> None:
         """Setup fixtures (ad-hoc)"""
-        pass
 
     @abstractmethod
     def run_case(self, testcase: DataDrivenTestCase) -> None:
