@@ -1,18 +1,18 @@
 """Tests for mypy incremental error output."""
-from typing import List
+from __future__ import annotations
 
 from mypy import build
-from mypy.test.helpers import assert_string_arrays_equal
-from mypy.test.data import DataDrivenTestCase, DataSuite
-from mypy.build import BuildSource
 from mypy.errors import CompileError
+from mypy.modulefinder import BuildSource
 from mypy.options import Options
+from mypy.test.data import DataDrivenTestCase, DataSuite
+from mypy.test.helpers import assert_string_arrays_equal
 
 
 class ErrorStreamSuite(DataSuite):
     required_out_section = True
-    base_path = '.'
-    files = ['errorstream.test']
+    base_path = "."
+    files = ["errorstream.test"]
 
     def run_case(self, testcase: DataDrivenTestCase) -> None:
         test_error_stream(testcase)
@@ -26,21 +26,19 @@ def test_error_stream(testcase: DataDrivenTestCase) -> None:
     options = Options()
     options.show_traceback = True
 
-    logged_messages = []  # type: List[str]
+    logged_messages: list[str] = []
 
-    def flush_errors(msgs: List[str], serious: bool) -> None:
+    def flush_errors(msgs: list[str], serious: bool) -> None:
         if msgs:
-            logged_messages.append('==== Errors flushed ====')
+            logged_messages.append("==== Errors flushed ====")
             logged_messages.extend(msgs)
 
-    sources = [BuildSource('main', '__main__', '\n'.join(testcase.input))]
+    sources = [BuildSource("main", "__main__", "\n".join(testcase.input))]
     try:
-        build.build(sources=sources,
-                    options=options,
-                    flush_errors=flush_errors)
+        build.build(sources=sources, options=options, flush_errors=flush_errors)
     except CompileError as e:
         assert e.messages == []
 
-    assert_string_arrays_equal(testcase.output, logged_messages,
-                               'Invalid output ({}, line {})'.format(
-                                   testcase.file, testcase.line))
+    assert_string_arrays_equal(
+        testcase.output, logged_messages, f"Invalid output ({testcase.file}, line {testcase.line})"
+    )
